/*
 * :set tabstop=4
 * efs_file.c
 * spd@daphne.cps.unizar.es
 * Sun Jun  8 22:54:26 DST 1997
 *
 * EFS file/dir high level I/O
 * IRIX 6.x and WinNT (ix86)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#ifdef WIN32
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>
#	include <malloc.h>

#	include "irix/types.h"
#	include "irix/param.h"
#	define	BITSPERBYTE	8

#	include "irix/dir.h"
#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"

#	ifndef S_ISDIR
#		define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR)
#	endif

#else
#	include <sys/dir.h>
#	include <sys/time.h>
#	include <sys/param.h>		/* DEV_BSIZE */
#	include <values.h>			/* BITSPERBYTE */

#	include <sys/dvh.h>			/* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>
#endif

#include <errno.h>
#include <time.h>

#include <assert.h>


#include "irix_efs.h"
#include "irix_io.h"
#include "efs_io.h"

#ifdef WIN32 
DWORD _n_read;
#endif

#define		MAXEXTENTLEN	(DEV_BSIZE * EFS_MAXEXTENTLEN)
#define DIRECTEXTENTS(di)	( (di)->di_numextents <= EFS_DIRECTEXTENTS )

EFS_FILE	_efs_fileb[EFS_MAX_FILE];

extern int read_extent( EFS_FS *fs, extent *ex, char *buffer);

static size_t get_next_extent(void *buffer, EFS_FILE *f);
static int load_extents( EFS_FILE *f );

/************************************************************************/

EFS_FILE *efs_fopen( char *name, char *type )
{
	register EFS_FILE	*p;
	register int		i;
	efs_ino_t			inum;
	unsigned char		buffer[DEV_BSIZE];
	EFS_FS				*fs;

	if ( NULL == (fs = efs_get_default_fs()))
		return 0;

	if ( 0 == ( inum=efs_path_to_inum(name)))
	{
		return(NULL);
	}


	for( p = &_efs_fileb[0], i=EFS_MAX_FILE; i-- ; p++)
	{
		if (!p->fs)
		{
			if ( NULL == (p->base = malloc((size_t)MAXEXTENTLEN)))
			{
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->di = malloc(sizeof(struct efs_dinode))))
			{
				free(p->base);
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->extents = malloc((size_t)MAXEXTENTLEN)))
			{
				free(p->base);
				free(p->di);
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}

			if ( 0 != efs_read_dinode( fs, inum, buffer, &(p->di) ))
			{
				efs_fclose( p );
				return(NULL);
			}
			p->fs =		fs;

			p->cnt =	0;
			p->ptr =	p->base;

			p->i =		0;
			p->block =	0;
			p->ie =		0;
			p->ed =		0;
			
			p->osize =	0;

			if ( !DIRECTEXTENTS(p->di) )
			{
				if ( ! load_extents( p ) )
				{
					free( p->extents );
					free( p->di );
					free( p->base );
					p->fs=NULL;
					return(NULL);
				}
			}
			return(p);
		}
	}
	efs_errno = EFS_EMFILE;
	return(NULL);
}

int efs_fclose( EFS_FILE *stream )
{
	if ( stream->extents )
		free( stream->extents );
	if ( stream->di )
		free( stream->di );
	if ( stream->base )
		free( stream->base );

	stream->fs=NULL;

	return(0);
}


size_t efs_fread(void *vptr, size_t size, size_t nitems, EFS_FILE *f)
{
	size_t			n, m, i;
	size_t			r=0;
	char			*ptr=vptr;

	if ( 0 == ( n = size * nitems ))
		return 0;
	/* disk eof test here */

	efs_errno = EFS_ENOERR;

	while ( n > 0 )
	{
		m = MIN( f->cnt, n );

		if ( m > 0 )
		{
			/*
			 * enough bytes in buffer
			 */
		
			if (ptr)
			{
				memcpy( ptr, f->ptr, m );
				ptr += m;
			}
			f->osize += m;
			f->ptr += m;
			f->cnt -= m;
			n -= m;
			r += m;
		}
		else
		{
			if ( -1 == ( i = get_next_extent(f->base, f)))
			{
				/* eof */
				f->osize ++;
				return r;
			}
			f->ptr = f->base;
			f->cnt = MIN( i, f->di->di_size - f->osize );
		}
		/* we have new data */
	}

	return (r);
}

/************************************************************************/

EFS_DIR *efs_opendir( char *name )
{
	struct stat	st;

	if ( -1 == efs_stat(name, &st))
		return(NULL);

	if ( S_ISDIR( st.st_mode ))
	{
		return( (EFS_DIR *) efs_fopen( name, "r"));
	}
	else
	{
		efs_errno = EFS_ENOTDIR;
		return(NULL);
	}
}

int efs_closedir( EFS_DIR *dir )
{
	return( efs_fclose ( (EFS_FILE *) dir ));
}

struct efs_direct *efs_readdir( EFS_DIR *f )
{
	struct efs_direct	*dirct=NULL;
	struct efs_dirblk	*dir;
	struct efs_dent		*ent;

	size_t			n;
	
	while ( dirct == NULL )
	{
		if ( f->cnt == 0)
		{
			if ( -1 == ( n = get_next_extent( f->base, f )))
			{
				/* eof */
				return NULL;
			}
			else
			{
				f->ptr = f->base;
				f->cnt = n;
				f->slot = 0;
				f->osize += n ;
			}
		}
		if ( f->cnt == 0)
		{	
			return NULL;
		}
		dir = (struct  efs_dirblk *)(f->ptr);

		if ( f->slot == dir->slots )
		{
			/* next dirblk */

			f->cnt -= sizeof( struct efs_dirblk );

			if ( f->cnt < 0 )
			{
				/* need another extent */
				f->cnt = 0;
				continue;
			}
			
			dir++ ;
			f->ptr = (unsigned char *)dir;
			
#ifdef LITTLEENDIAN
			SWAP_WORD(dir->magic);
#endif /* LITTLEENDIAN */

			if (dir->magic != EFS_DIRBLK_MAGIC)
				return NULL;

			f->slot = 0;

			if ( dir->slots == 0 )
				continue;
		}

		if ( dir->space[f->slot] != 0 )
		{

			ent = EFS_SLOTOFF_TO_DEP(
				dir,
				EFS_REALOFF(dir->space[f->slot])
				);
#ifdef LITTLEENDIAN
			SWAP_HALF(ent->ud_inum.s[0]);
			SWAP_HALF(ent->ud_inum.s[1]);
#endif /* LITTLEENDIAN */
			strncpy( f->dct.d_name, ent->d_name, ent->d_namelen);
			f->dct.d_name[ent->d_namelen] = 0x00 ;
			f->dct.d_namlen = ent->d_namelen;
			f->dct.d_ino = EFS_GET_INUM(ent);
/*			f->dct.d_ino = ent->ud_inum.s[1];*/
			f->dct.d_reclen = DIRSIZ(&(f->dct));
			
			dirct=&f->dct;

		}
		f->slot ++;
	}

	return dirct;
}

/************************************************************************/

/************************************************************************/

static size_t get_next_extent(void *buffer, EFS_FILE *f)
{

	/*
	 * Get next fs block
	 */
	extent			*ex=NULL;

	if ( DIRECTEXTENTS(f->di) )
	{
		if ( f->i == f->di->di_numextents )
		{
			/* done */
			return -1;
		}
		else
		{
			ex = &f->di->di_u.di_extents[f->i];
			f->i ++;
		}	
	}
	else
	{
		if ( f->ed == f->di->di_numextents )
		{
			/* done */
			return -1;
		}
		else
		{
			if ( f->ie == (512 / sizeof( extent)) )
			{
				/* I need another indirect block */
				f->ie = 0;
				f->block++;

				if ( f->block == f->di->di_u.di_extents[f->i].ex_length )
				{
					/* I need another extent from original dinode */

					f->i++;
					if ( f->i == f->di->di_u.di_extents[0].ex_offset )
					{
						/*
						 * hmmm... this should not be reached,
						 * since f->ed is tested above...
						 */
						return -1;
						/* We are done! */
					}
					
					/* i is valid, start with blocks */
					f->block = 0;
				}
				load_extents(f);
			}
			/* so now ie points to valid data (or eof) */

			ex = &f->extents[f->ie];
			f->ie++;
			f->ed++;

		}
	}

	/* ex points to valid data (or eof) */

	/* read data */

	assert( f->osize == 512 * ex->ex_offset );
	return (read_extent( f->fs, ex, buffer));
}

/************************************************************************/


static int load_extents( EFS_FILE *f )
{
	off_t			start;

	start = 512 * (
			f->fs->vh->vh_pt[f->fs->partition].pt_firstlbn +
			f->di->di_u.di_extents[f->i].ex_bn + f->block );

	if ( start == LSEEK(f->fs->file, start, SEEK_SET))
	{
		if ( 512 == READ( f->fs->file, f->extents, 512))
		{
#ifdef LITTLEENDIAN
			{
				extent *foo = (extent *)f->extents;
				unsigned int ui;

				while( foo < ((extent *)f->extents)+ 512 / sizeof(extent) )
				{
					ui=(unsigned int)*((unsigned int *) foo);
					SWAP_WORD(ui);
					foo->ex_magic = ui >> 24;
					foo->ex_bn = ( ui & 0x00FFFFFF );

					ui=(unsigned int)*((unsigned int *) foo + 1);
					SWAP_WORD(ui);
					foo->ex_length = ui >> 24;
					foo->ex_offset = ( ui & 0x00FFFFFF );
					foo++;
				}

			}
#endif /* LITTLEENDIAN */
		}
		else
			efs_errno = EFS_EREAD;
		}
	else
		efs_errno = EFS_ESEEK;

	return ( efs_errno == EFS_ENOERR );
}
