/*
 * :set tabstop=4
 * efs_fs.c
 * spd@daphne.cps.unizar.es
 * Wed Jun  4 22:34:20 DST 1997
 *
 * efs mounting
 * IRIX 6.x, WinNT
 *
 * All fs related calculations are macro-based, so you don't need
 * external libraries
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef WIN32
#	include <sys/dir.h>
#	include <sys/time.h>
#	include <sys/param.h>       /* DEV_BSIZE */
#	include <values.h>          /* BITSPERBYTE */

#	include <sys/dvh.h>         /* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>
#else
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>
#	include <malloc.h>

#	include "irix/types.h"
#	include "irix/param.h"
#	define	BITSPERBYTE	8

#	include "irix/dir.h"
#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"

#	ifndef S_ISDIR
#		define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR)
#	endif
#endif

#include "irix_efs.h"
#include "irix_io.h"
#include "efs_io.h"

/*
 * little ego
 */
static char *__ident="@(#)libefs 0.0.1 (c) SPDsoft, Sep 30, 1997";

/*
 * ANSI:  If no explicit initialization is performed, static
 * objects are implicitly initialized to zero. 
 */


EFS_DISK	_efs_diskb[EFS_MAX_DISKS];
EFS_FS		_efs_fsb[EFS_MAX_FSS];
static EFS_FS	*_current_fs=NULL;

int  efs_errno=EFS_ENOERR;
char *efs_sys_errlist[]=
	{
		"No error",
		"Bad checksum",
		"Bad magic number",
		"Read error",
		"Empty partition",
		"Bad partition",
		"Seek error",
		"Resource temporarily unavailable",
		"Too many open files",
		"Not a directory",
		"Filename too long",
		"No such file or directory"
	};

static int check_dk_vh(struct volume_header *vh);
static int check_sblock(struct efs *efs);


#ifdef WIN32 
DWORD _n_read;
#endif


EFS_DISK *efs_open_disk( char *name )
{
	register EFS_DISK *p;
	register int i;

	for( p = &_efs_diskb[0], i=EFS_MAX_DISKS; i-- ; p++)
	{
		if (!p->file)
		{
			if ( NULL == (p->vh = malloc(DEV_BSIZE)))
			{
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}
#ifdef WIN32 
			if (INVALID_HANDLE_VALUE == ( p->file = CreateFile(
			(LPCTSTR) name,
			GENERIC_READ,
			0, /* Not shared */
			NULL, /* Not inherited by child processes */
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
			NULL
			)))
#else
			if ( -1 == ( p->file = open( name, O_RDONLY )))
#endif
			{
				p->file=(EFS_DISK_T)0L;
				return(NULL);
			}
			else
			{
				if ( DEV_BSIZE != READ( p->file, (void *)p->vh, DEV_BSIZE ))
				{
					efs_errno = EFS_EREAD;
					efs_close_disk( p );
					return (NULL);
				}
				else
				{
					if (!check_dk_vh( p->vh ))
					{
						efs_close_disk( p );
						return (NULL);
					}
				}
				return(p);
			}
		}
	}
	efs_errno = EFS_EMFILE;
	return(NULL);
}

int efs_close_disk( EFS_DISK *disk )
{
	int		err;

	err = CLOSE( disk->file );

	disk->file=(EFS_DISK_T)0L;
	if (disk->vh)
		free( disk->vh );
	
	return(err);
}

/************************************************************************/

EFS_FS   *efs_mount( EFS_DISK *disk, int partition, int flags )
{
	register EFS_FS		*p;
	register int		i;
	OFF_T				start;
	unsigned char		sblock[DEV_BSIZE]; /* superblock */


	if ( efs_empty_partition( disk, partition ))
	{
		efs_errno = EFS_EEMPTY;
		return (NULL);
	}

	if (
		(!( flags & EFS_FMOUNTALL ) && (
			( disk->vh->vh_pt[partition].pt_type != PTYPE_SYSV ) &&
			( disk->vh->vh_pt[partition].pt_type != PTYPE_EFS )
		))||
		( partition < 0 ) || ( partition >= NPARTAB )
	)
	{
		efs_errno = EFS_EBADPART;
		return (NULL);
	}

	for( p = &_efs_fsb[0], i=EFS_MAX_FSS; i-- ; p++)
	{
		if (!p->file)
		{
			if ( NULL == (p->efs = malloc(sizeof(struct efs))))
			{
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->cwd = malloc(MAXPATHLEN)))
			{
				free(p->efs);
				efs_errno=EFS_EAGAIN;
				return(NULL);
			}
			p->file = disk->file;
			p->partition = partition;
			p->vh = disk->vh;

			p->inum = EFS_ROOTINO;
			strcpy(p->cwd, "/");

			/*
			 * Translate partition starting cylinder to offset from 0
			 */

			start =
				( 1 + p->vh->vh_pt[partition].pt_firstlbn ) * 512 ;

			if ( start != LSEEK(p->file, start, SEEK_SET))
			{
				efs_errno = EFS_ESEEK;
				efs_umount( p );
				return (NULL);
			}
			else if ( sizeof(sblock) !=
				READ( p->file, (void *)sblock, sizeof(sblock) ))
			{
				efs_errno = EFS_EREAD;
				efs_umount( p );
				return (NULL);
			}
			else
			{
				memcpy( p->efs, sblock, sizeof(struct efs));
				if (!check_sblock( p->efs ))
				{
					efs_errno = EFS_EMAGIC;
					efs_umount( p );
					return (NULL);
				}
			}
			if ( _current_fs==NULL )
				_current_fs=p;
			return(p);
		}
	}
	efs_errno = EFS_EMFILE;
	return(NULL);
}

int  efs_umount(EFS_FS *fs)
{
	fs->file=(EFS_DISK_T)0L;
	if (fs->efs)
		free( fs->efs );
	if (fs->cwd)
		free( fs->cwd );

	if ( _current_fs == fs )
		_current_fs = NULL;

	return(0);
}

EFS_FS   *efs_get_default_fs( void )
{
	return _current_fs;
}
void efs_set_default_fs( EFS_FS *fs )
{
	_current_fs = fs;
}

char *efs_getwd(char *pathname)
{
	if ( !_current_fs )
	{
		strcpy(pathname, "No current fs");
		return(NULL);
	}
	else
	{
		strncpy( pathname, _current_fs->cwd, MAXPATHLEN);
		return( pathname );
	}
}

int      efs_chdir(char *ipath)
{
	efs_ino_t		inum;
	struct stat		st;

	char		path[MAXPATHLEN];
	char		ncwd[MAXPATHLEN];
	char		*argt;
	char		*p;
	int			rem;

	if ( 0 == (inum = efs_path_to_inum(ipath)))
		return(-1);

	if ( -1 == efs_istat(inum, &st))
		return(-1);
	
	if ( S_ISDIR( st.st_mode ))
	{
		strncpy(path, ipath, MAXPATHLEN);
/*
 * assertion:
 *		cwd : "/" or ("/name")* where name != "." or ".."
 */
		if ( *path == '/' )
			ncwd[0]=0x00;
		else
		{
			if ( 0==strcmp( _current_fs->cwd, "/") )
				ncwd[0]=0x00;
			else
				strncpy(ncwd, _current_fs->cwd, MAXPATHLEN);
		}
		rem = MAXPATHLEN - strlen(ncwd);
	
		for(
			argt = strtok( path, "/" );
			argt != NULL;
			argt = strtok( NULL, "/" )
		)
		{
			if ( 0==strcmp( argt, ".") )
				continue;
		
			if ( 0==strcmp( argt, ".."))
			{
				if ( p = (strrchr(ncwd, (int)'/')))
				{
					*p = 0x00;
					rem = MAXPATHLEN - strlen(ncwd);
				}
			}
			else
			{
				rem -= 1 + strlen(argt);
				if ( rem > 0 )
				{
					strcat( ncwd, "/" );
					strcat( ncwd, argt );
				}
				else
				{
					efs_errno = EFS_ENAMETOOLONG;
					return -1;
				}
			}
		}
		if ( *ncwd == 0x00 )
			strcpy(_current_fs->cwd, "/");
		else
			strncpy(_current_fs->cwd, ncwd, MAXPATHLEN);

		_current_fs->inum = inum;
		return 0;
	}
	else
	{
		efs_errno=EFS_ENOTDIR;
		return (-1);
	}
}


/************************************************************************/
/************************************************************************/

static int check_dk_vh(struct volume_header *vh)
{
	short		i;
	int			sum, *word;

/*	assertion:  ( sizeof(int) == 4 ) */

	for (
		word = (int *)vh, sum=0;
		word < ((int *)vh + sizeof(*vh)/sizeof(int) -2);
		word ++
	)
#ifdef LITTLEENDIAN
	sum += swap_word(*word);
#else
	sum += *word;
#endif /* LITTLEENDIAN */


#ifdef LITTLEENDIAN
	SWAP_WORD(vh->vh_magic);
	SWAP_HALF(vh->vh_rootpt);
	SWAP_HALF(vh->vh_swappt);
	SWAP_HALF(vh->vh_dp.dp_cyls);
	SWAP_HALF(vh->vh_dp.dp_shd0);
	SWAP_HALF(vh->vh_dp.dp_trks0);
	SWAP_HALF(vh->vh_dp.dp_unused);
	SWAP_HALF(vh->vh_dp.dp_secs);
	SWAP_HALF(vh->vh_dp.dp_secbytes);
	SWAP_HALF(vh->vh_dp.dp_interleave);
	SWAP_WORD(vh->vh_dp.dp_flags);
	SWAP_WORD(vh->vh_dp.dp_datarate);
	SWAP_WORD(vh->vh_dp.dp_nretries);
	SWAP_WORD(vh->vh_dp.dp_mspw);
	SWAP_HALF(vh->vh_dp.dp_xgap1);
	SWAP_HALF(vh->vh_dp.dp_xsync);
	SWAP_HALF(vh->vh_dp.dp_xrdly);
	SWAP_HALF(vh->vh_dp.dp_xgap2);
	SWAP_HALF(vh->vh_dp.dp_xrgate);
	SWAP_HALF(vh->vh_dp.dp_xwcont);
	for(i=0; i<NVDIR; i++)
	{
		SWAP_WORD(vh->vh_vd[i].vd_lbn);
		SWAP_WORD(vh->vh_vd[i].vd_nbytes);
	}
	for(i=0; i<NPARTAB; i++)
	{
		SWAP_WORD(vh->vh_pt[i].pt_nblks);
		SWAP_WORD(vh->vh_pt[i].pt_firstlbn);
		SWAP_WORD(vh->vh_pt[i].pt_type);
	}
	SWAP_WORD(vh->vh_fill);
	SWAP_WORD(vh->vh_csum);
#endif /* LITTLEENDIAN */

	efs_errno = EFS_ENOERR;
	if ( vh->vh_csum != -sum )
		efs_errno = EFS_ECSUM;
	if ( vh->vh_magic != VHMAGIC )
		efs_errno = EFS_EMAGIC;

	return ( efs_errno == EFS_ENOERR );
}
/************************************************************************/

static int		check_sblock(struct efs *efs)
{
/*
 * This is the number of longs contained in the region of the
 * superblock that we need to scan in order to compute the checksum.
 */
#define SBLCKSZ 22

	int		i;
	long	t6;
	long	t7;
	long	t8 = 0;
	long	v1 = 0;
	long	*lptr;
	lptr	= (long *)efs;

	/*
	 * Checksum by wpaul@ctr.columbia.edu (Bill Paul)
	 */

    /*
	 * We pass over the superblock data 16 bits at a time.
	 * We XOR each set of 16 bits into an accumulator (v1).
	 * We then shift the accumulator one bit to the left and store
	 * the result into a temporary value (t7). We then store
	 * the high bit of the (unshifted) accumulator in the low
	 * bit of a second temporary value (t8). The accumulator
	 * is then set equal to the logical OR of t7 and t8. Repeat
	 * until you run out of bits. At the end, the accumulator
	 * is the checksum.
	 */

    i = SBLCKSZ;
    while(i)
	{
#ifndef LITTLEENDIAN
		/* Process upper 16 bits */
		t6 = (*lptr & 0xFFFF0000) >> 16;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;

		/* Process lower 16 bits */
		t6 = *lptr & 0x0000FFFF;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;
#else
		long l;

		l=swap_word(*lptr);
		
		/* Process upper 16 bits */
		t6 = (l & 0xFFFF0000) >> 16;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;

		/* Process lower 16 bits */
		t6 = l & 0x0000FFFF;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;
#endif
		lptr++;
		i--;
	}

#ifdef LITTLEENDIAN
	SWAP_WORD(efs->fs_size);
	SWAP_WORD(efs->fs_firstcg);
	SWAP_WORD(efs->fs_cgfsize);
	SWAP_HALF(efs->fs_cgisize);
	SWAP_HALF(efs->fs_sectors);
	SWAP_HALF(efs->fs_heads);
	SWAP_HALF(efs->fs_ncg);
	SWAP_HALF(efs->fs_dirty);
	SWAP_WORD(efs->fs_time);
	SWAP_WORD(efs->fs_magic);

	SWAP_WORD(efs->fs_bmsize);
	SWAP_WORD(efs->fs_tfree);
	SWAP_WORD(efs->fs_tinode);
	SWAP_WORD(efs->fs_bmblock);
	SWAP_WORD(efs->fs_replsb);
	SWAP_WORD(efs->fs_lastialloc);

	SWAP_WORD(efs->fs_checksum);
#endif /* LITTLEENDIAN */

	return (
		(efs->fs_magic == EFS_MAGIC) &&
		(efs->fs_checksum == v1) &&
		(efs->fs_size > 0) &&
		(efs->fs_ncg > 0) &&
		(efs->fs_firstcg > 0) &&
		(efs->fs_cgfsize > 0) &&
		(efs->fs_cgisize > 0) &&
		(efs->fs_tfree > 0) &&
		(efs->fs_tinode > 0)
	);
}
